from __future__ import annotations

__all__ = (
    "fix_package_names",
    "install_lazy_importer",
    "set_deprecated_aliases",
)

import ast
import inspect
import sys
import warnings
from importlib import import_module
from types import ModuleType
from typing import Any


def install_lazy_importer() -> bool:
    module_globals = sys._getframe(1).f_globals
    module_name = module_globals["__name__"]
    module_prefix = module_name + "."
    module = sys.modules[module_name]
    lazy_map, deprecated_aliases = _build_lazy_map(module)
    names = sorted(lazy_map)

    # Delete symbols that are not part of the API
    del module_globals["TYPE_CHECKING"]
    del module_globals["install_lazy_importer"]

    if not lazy_map and not deprecated_aliases:
        return False

    def __getattr__(name: str) -> Any:
        if new_name := deprecated_aliases.get(name):
            emit_deprecation_warning(module_name, name, new_name)
            target_mod, target_attr = new_name.rsplit(".", 1)
        else:
            try:
                target_mod, target_attr = lazy_map[name]
            except KeyError:
                raise AttributeError(
                    f"module {module_name!r} has no attribute {name!r}"
                ) from None

        imported = import_module(target_mod, module_name)
        value = getattr(imported, target_attr)

        # patch the module name to match
        if (
            getattr(value, "__module__", "").startswith(module_prefix)
            and name not in deprecated_aliases
        ):
            value.__module__ = module_name

        module_globals[name] = value
        return value

    def __dir__() -> list[str]:
        return names

    module_globals["__dir__"] = __dir__
    module_globals["__getattr__"] = __getattr__
    module_globals.pop("fix_package_names", None)
    module_globals.pop("set_deprecated_aliases", None)
    return True


def fix_package_names() -> None:
    module_globals = sys._getframe(1).f_globals
    module_prefix = module_globals["__name__"] + "."
    del module_globals[fix_package_names.__name__]
    for value in module_globals.values():
        if modname := getattr(value, "__module__", ""):
            if modname.startswith(module_prefix):
                parts = modname.split(".")
                value.__module__ = ".".join(
                    part for part in parts if not part.startswith("_")
                )


def emit_deprecation_warning(module_name: str, name: str, target: str) -> None:
    warnings.warn(
        f"The {module_name}.{name} alias is deprecated, use {target} instead.",
        DeprecationWarning,
        stacklevel=3,
    )


def set_deprecated_aliases(aliases: dict[str, str]) -> None:
    module_globals = sys._getframe(1).f_globals
    module_name = module_globals["__name__"]
    del module_globals[set_deprecated_aliases.__name__]

    def __getattr__(name: str) -> Any:
        try:
            target = aliases[name]
        except KeyError:
            raise AttributeError(
                f"module {module_name!r} has no attribute {name!r}"
            ) from None

        emit_deprecation_warning(module_name, name, target)
        target_modname, attrname = target.rsplit(".", 1)
        module = import_module(target_modname)
        return getattr(module, attrname)

    sys.modules[module_name].__dict__["__getattr__"] = __getattr__


def _build_lazy_map(
    module: ModuleType,
) -> tuple[dict[str, tuple[str, str]], dict[str, str]]:
    try:
        source = inspect.getsource(module)
    except OSError:
        return {}, {}

    tree = compile(source, module.__file__ or "", "exec", ast.PyCF_ONLY_AST)
    assert isinstance(tree, ast.Module)
    out: dict[str, tuple[str, str]] = {}
    deprecated_aliases: dict[str, str] = {}

    for node in tree.body:
        if not isinstance(node, ast.If) or not _is_type_checking_block(node.test):
            continue

        for stmt in node.body:
            match stmt:
                case ast.ImportFrom():
                    base = "." * stmt.level + (stmt.module or "")
                    for alias in stmt.names:
                        if alias.name == "*":
                            raise RuntimeError("star imports not supported")

                        exported = alias.asname or alias.name
                        out[exported] = (base, alias.name)
                case ast.Expr() if isinstance(stmt.value, ast.Call):
                    call = stmt.value
                    if (
                        isinstance(call.func, ast.Name)
                        and call.func.id == "set_deprecated_aliases"
                    ):
                        arg0 = call.args[0]
                        assert isinstance(arg0, ast.Dict)
                        for key, value in zip(arg0.keys, arg0.values, strict=True):
                            assert isinstance(key, ast.Constant)
                            assert isinstance(key.value, str)
                            assert isinstance(value, ast.Constant)
                            assert isinstance(value.value, str)
                            deprecated_aliases[key.value] = value.value

    return out, deprecated_aliases


def _is_type_checking_block(test: ast.AST) -> bool:
    if not isinstance(test, ast.BoolOp):
        return False

    subtest = test.values[0]
    match subtest:
        case ast.Name():
            return subtest.id == "TYPE_CHECKING"
        case ast.Attribute():
            return (
                isinstance(subtest.value, ast.Name)
                and subtest.value.id == "typing"
                and subtest.attr == "TYPE_CHECKING"
            )
        case _:
            return False



from __future__ import annotations

from typing import TYPE_CHECKING

from .._lazyimport import (
    fix_package_names,
    install_lazy_importer,
    set_deprecated_aliases,
)

if TYPE_CHECKING or not install_lazy_importer():
    from ._eventloop import AsyncBackend as AsyncBackend
    from ._resources import AsyncResource as AsyncResource
    from ._sockets import ConnectedUDPSocket as ConnectedUDPSocket
    from ._sockets import ConnectedUNIXDatagramSocket as ConnectedUNIXDatagramSocket
    from ._sockets import IPAddressType as IPAddressType
    from ._sockets import IPSockAddrType as IPSockAddrType
    from ._sockets import SocketAttribute as SocketAttribute
    from ._sockets import SocketListener as SocketListener
    from ._sockets import SocketStream as SocketStream
    from ._sockets import UDPPacketType as UDPPacketType
    from ._sockets import UDPSocket as UDPSocket
    from ._sockets import UNIXDatagramPacketType as UNIXDatagramPacketType
    from ._sockets import UNIXDatagramSocket as UNIXDatagramSocket
    from ._sockets import UNIXSocketStream as UNIXSocketStream
    from ._streams import AnyByteReceiveStream as AnyByteReceiveStream
    from ._streams import AnyByteSendStream as AnyByteSendStream
    from ._streams import AnyByteStream as AnyByteStream
    from ._streams import AnyByteStreamConnectable as AnyByteStreamConnectable
    from ._streams import (
        AnyUnreliableByteReceiveStream as AnyUnreliableByteReceiveStream,
    )
    from ._streams import AnyUnreliableByteSendStream as AnyUnreliableByteSendStream
    from ._streams import AnyUnreliableByteStream as AnyUnreliableByteStream
    from ._streams import ByteReceiveStream as ByteReceiveStream
    from ._streams import ByteSendStream as ByteSendStream
    from ._streams import ByteStream as ByteStream
    from ._streams import ByteStreamConnectable as ByteStreamConnectable
    from ._streams import Listener as Listener
    from ._streams import ObjectReceiveStream as ObjectReceiveStream
    from ._streams import ObjectSendStream as ObjectSendStream
    from ._streams import ObjectStream as ObjectStream
    from ._streams import ObjectStreamConnectable as ObjectStreamConnectable
    from ._streams import UnreliableObjectReceiveStream as UnreliableObjectReceiveStream
    from ._streams import UnreliableObjectSendStream as UnreliableObjectSendStream
    from ._streams import UnreliableObjectStream as UnreliableObjectStream
    from ._subprocesses import Process as Process
    from ._tasks import TaskGroup as TaskGroup
    from ._tasks import TaskStatus as TaskStatus
    from ._testing import TestRunner as TestRunner

    fix_package_names()
    set_deprecated_aliases(
        {
            "CapacityLimiter": "anyio.CapacityLimiter",
            "Condition": "anyio.Condition",
            "Event": "anyio.Event",
            "Lock": "anyio.Lock",
            "Semaphore": "anyio.Semaphore",
            "CancelScope": "anyio.CancelScope",
            "BlockingPortal": "anyio.from_thread.BlockingPortal",
        }
    )



from __future__ import annotations

import dataclasses
import socket
import sys
from collections.abc import Callable, Generator, Iterator
from contextlib import ExitStack, contextmanager
from inspect import isasyncgenfunction, iscoroutinefunction, ismethod
from typing import TYPE_CHECKING, Any, cast

import pytest
from _pytest.fixtures import FuncFixtureInfo, SubRequest
from _pytest.outcomes import Exit
from _pytest.scope import Scope

from . import get_available_backends
from ._core._eventloop import (
    current_async_library,
    get_async_backend,
    reset_current_async_library,
    set_current_async_library,
)
from ._core._exceptions import iterate_exceptions

if TYPE_CHECKING:
    from .abc import TestRunner

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

if TYPE_CHECKING:
    # pytest >= 9.2 keeps CallSpec2 as a TYPE_CHECKING-only alias of CallSpec
    from _pytest.python import CallSpec2 as CallSpec
else:
    try:
        # c.f. https://github.com/pytest-dev/pytest/pull/14742
        # pytest >= 9.2
        from _pytest.python import CallSpec
    except ImportError:
        from _pytest.python import CallSpec2 as CallSpec

_current_runner: TestRunner | None = None
_runner_stack: ExitStack | None = None
_runner_leases = 0


def extract_backend_and_options(backend: object) -> tuple[str, dict[str, Any]]:
    if isinstance(backend, str):
        return backend, {}
    elif isinstance(backend, tuple) and len(backend) == 2:
        if isinstance(backend[0], str) and isinstance(backend[1], dict):
            return cast(tuple[str, dict[str, Any]], backend)

    raise TypeError("anyio_backend must be either a string or tuple of (string, dict)")


@contextmanager
def get_runner(
    backend_name: str, backend_options: dict[str, Any]
) -> Iterator[TestRunner]:
    global _current_runner, _runner_leases, _runner_stack
    if _current_runner is None:
        asynclib = get_async_backend(backend_name)
        _runner_stack = ExitStack()
        if current_async_library() is None:
            # Since we're in control of the event loop, we can cache the name of the
            # async library
            token = set_current_async_library(backend_name)
            _runner_stack.callback(reset_current_async_library, token)

        backend_options = backend_options or {}
        _current_runner = _runner_stack.enter_context(
            asynclib.create_test_runner(backend_options)
        )

    _runner_leases += 1
    try:
        yield _current_runner
    finally:
        _runner_leases -= 1
        if not _runner_leases:
            assert _runner_stack is not None
            _runner_stack.close()
            _runner_stack = _current_runner = None


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("anyio")
    parser.addini(
        "anyio_mode",
        default="strict",
        help='AnyIO plugin mode (either "strict" or "auto")',
    )
    group.addoption(
        "--anyio-mode",
        default=None,
        dest="anyio_mode",
        help="""
        'auto'   - All async test functions will be handled by AnyIO pytest plugin
        'strict' - Disabling autoprocessing(useful when anyio tests
            need to coexist with other async test plugins)
        """,
    )


def _is_auto_mode(config: pytest.Config, mode: str) -> bool:
    return (
        config.getoption(mode, default=None) == "auto" or config.getini(mode) == "auto"
    )


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        "markers",
        "anyio: mark the (coroutine function) test to be run asynchronously via anyio.",
    )
    if (
        _is_auto_mode(config, "anyio_mode")
        and config.pluginmanager.has_plugin("asyncio")
        and _is_auto_mode(config, "asyncio_mode")
    ):
        config.issue_config_time_warning(
            pytest.PytestConfigWarning(
                "AnyIO auto mode has been enabled together with pytest-asyncio auto "
                "mode. This may cause unexpected behavior."
            ),
            1,
        )


@pytest.hookimpl(hookwrapper=True)
def pytest_fixture_setup(fixturedef: Any, request: Any) -> Generator[Any]:
    def wrapper(anyio_backend: Any, request: SubRequest, **kwargs: Any) -> Any:
        # Rebind any fixture methods to the request instance
        if (
            request.instance
            and ismethod(func)
            and type(func.__self__) is type(request.instance)
        ):
            local_func = func.__func__.__get__(request.instance)
        else:
            local_func = func

        backend_name, backend_options = extract_backend_and_options(anyio_backend)
        if has_backend_arg:
            kwargs["anyio_backend"] = anyio_backend

        if has_request_arg:
            kwargs["request"] = request

        with get_runner(backend_name, backend_options) as runner:
            # re-entrant call into the test runner detected. this happens when an async fixture
            # is dynamically requested via request.getfixturevalue() from inside a running async
            # test or fixture. on asyncio this raises RuntimeError: This event loop is already
            # running, on trio the runner deadlocks - the host loop blocks waiting for the
            # coroutine to return, but the coroutine is waiting for the host loop. raising here
            # prevents the hang and gives a consistent error across backends.
            if runner.is_running():
                raise RuntimeError(
                    "Cannot schedule a coroutine in the test runner while another is already running; "
                    "likely caused by request.getfixturevalue() on an async fixture."
                )

            if isasyncgenfunction(local_func):
                yield from runner.run_asyncgen_fixture(local_func, kwargs)
            else:
                yield runner.run_fixture(local_func, kwargs)

    # Only apply this to coroutine functions and async generator functions in requests
    # that involve the anyio_backend fixture
    func = fixturedef.func
    if isasyncgenfunction(func) or iscoroutinefunction(func):
        if "anyio_backend" in request.fixturenames:
            fixturedef.func = wrapper
            original_argname = fixturedef.argnames

            if not (has_backend_arg := "anyio_backend" in fixturedef.argnames):
                fixturedef.argnames += ("anyio_backend",)

            if not (has_request_arg := "request" in fixturedef.argnames):
                fixturedef.argnames += ("request",)

            try:
                return (yield)
            finally:
                fixturedef.func = func
                fixturedef.argnames = original_argname

    return (yield)


@pytest.hookimpl(tryfirst=True)
def pytest_pycollect_makeitem(
    collector: pytest.Module | pytest.Class, name: str, obj: object
) -> None:
    if collector.istestfunction(obj, name):
        inner_func = obj.hypothesis.inner_test if hasattr(obj, "hypothesis") else obj
        if iscoroutinefunction(inner_func):
            anyio_auto_mode = _is_auto_mode(collector.config, "anyio_mode")
            marker = collector.get_closest_marker("anyio")
            own_markers = getattr(obj, "pytestmark", ())
            if (
                anyio_auto_mode
                or marker
                or any(marker.name == "anyio" for marker in own_markers)
            ):
                pytest.mark.usefixtures("anyio_backend")(obj)


def pytest_collection_finish(session: pytest.Session) -> None:
    for i, item in reversed(list(enumerate(session.items))):
        if (
            isinstance(item, pytest.Function)
            and iscoroutinefunction(item.function)
            and item.get_closest_marker("anyio") is not None
            and "anyio_backend" not in item.fixturenames
        ):
            new_items = []
            try:
                cs_fields = {f.name for f in dataclasses.fields(CallSpec)}
            except TypeError:
                cs_fields = set()

            for param_index, backend in enumerate(get_available_backends()):
                if "_arg2scope" in cs_fields:  # pytest >= 8
                    callspec = CallSpec(
                        params={"anyio_backend": backend},
                        indices={"anyio_backend": param_index},
                        _arg2scope={"anyio_backend": Scope.Module},
                        _idlist=[backend],
                        marks=[],
                    )
                else:  # pytest 7.x
                    callspec = CallSpec(  # type: ignore[call-arg]
                        funcargs={},
                        params={"anyio_backend": backend},
                        indices={"anyio_backend": param_index},
                        arg2scope={"anyio_backend": Scope.Module},
                        idlist=[backend],
                        marks=[],
                    )

                fi = item._fixtureinfo
                new_names_closure = list(fi.names_closure)
                if "anyio_backend" not in new_names_closure:
                    new_names_closure.append("anyio_backend")

                new_fixtureinfo = FuncFixtureInfo(
                    argnames=fi.argnames,
                    initialnames=fi.initialnames,
                    names_closure=new_names_closure,
                    name2fixturedefs=fi.name2fixturedefs,
                )
                new_item = pytest.Function.from_parent(
                    item.parent,
                    name=f"{item.originalname}[{backend}]",
                    callspec=callspec,
                    callobj=item.obj,
                    fixtureinfo=new_fixtureinfo,
                    keywords=item.keywords,
                    originalname=item.originalname,
                )
                new_items.append(new_item)

            session.items[i : i + 1] = new_items


@pytest.hookimpl(tryfirst=True)
def pytest_pyfunc_call(pyfuncitem: Any) -> bool | None:
    def run_with_hypothesis(**kwargs: Any) -> None:
        with get_runner(backend_name, backend_options) as runner:
            runner.run_test(original_func, kwargs)

    backend = pyfuncitem.funcargs.get("anyio_backend")
    if backend:
        backend_name, backend_options = extract_backend_and_options(backend)

        if hasattr(pyfuncitem.obj, "hypothesis"):
            # Wrap the inner test function unless it's already wrapped
            original_func = pyfuncitem.obj.hypothesis.inner_test
            if original_func.__qualname__ != run_with_hypothesis.__qualname__:
                if iscoroutinefunction(original_func):
                    pyfuncitem.obj.hypothesis.inner_test = run_with_hypothesis

            return None

        if iscoroutinefunction(pyfuncitem.obj):
            funcargs = pyfuncitem.funcargs
            testargs = {arg: funcargs[arg] for arg in pyfuncitem._fixtureinfo.argnames}
            with get_runner(backend_name, backend_options) as runner:
                try:
                    runner.run_test(pyfuncitem.obj, testargs)
                except ExceptionGroup as excgrp:
                    for exc in iterate_exceptions(excgrp):
                        if isinstance(exc, (Exit, KeyboardInterrupt, SystemExit)):
                            raise exc from excgrp

                    raise

            return True

    return None


@pytest.fixture(scope="module", params=get_available_backends())
def anyio_backend(request: Any) -> Any:
    return request.param


@pytest.fixture
def anyio_backend_name(anyio_backend: Any) -> str:
    if isinstance(anyio_backend, str):
        return anyio_backend
    else:
        return anyio_backend[0]


@pytest.fixture
def anyio_backend_options(anyio_backend: Any) -> dict[str, Any]:
    if isinstance(anyio_backend, str):
        return {}
    else:
        return anyio_backend[1]


class FreePortFactory:
    """
    Manages port generation based on specified socket kind, ensuring no duplicate
    ports are generated.

    This class provides functionality for generating available free ports on the
    system. It is initialized with a specific socket kind and can generate ports
    for given address families while avoiding reuse of previously generated ports.

    Users should not instantiate this class directly, but use the
    ``free_tcp_port_factory`` and ``free_udp_port_factory`` fixtures instead. For simple
    uses cases, ``free_tcp_port`` and ``free_udp_port`` can be used instead.
    """

    def __init__(self, kind: socket.SocketKind) -> None:
        self._kind = kind
        self._generated = set[int]()

    @property
    def kind(self) -> socket.SocketKind:
        """
        The type of socket connection (e.g., :data:`~socket.SOCK_STREAM` or
        :data:`~socket.SOCK_DGRAM`) used to bind for checking port availability

        """
        return self._kind

    def __call__(self, family: socket.AddressFamily | None = None) -> int:
        """
        Return an unbound port for the given address family.

        :param family: if omitted, both IPv4 and IPv6 addresses will be tried
        :return: a port number

        """
        if family is not None:
            families = [family]
        else:
            families = [socket.AF_INET]
            if socket.has_ipv6:
                families.append(socket.AF_INET6)

        while True:
            port = 0
            with ExitStack() as stack:
                for fam in families:
                    sock = stack.enter_context(socket.socket(fam, self._kind))
                    addr = "::1" if fam == socket.AF_INET6 else "127.0.0.1"
                    try:
                        sock.bind((addr, port))
                    except OSError:
                        break

                    if not port:
                        port = sock.getsockname()[1]
                else:
                    if port not in self._generated:
                        self._generated.add(port)
                        return port


@pytest.fixture(scope="session")
def free_tcp_port_factory() -> FreePortFactory:
    return FreePortFactory(socket.SOCK_STREAM)


@pytest.fixture(scope="session")
def free_udp_port_factory() -> FreePortFactory:
    return FreePortFactory(socket.SOCK_DGRAM)


@pytest.fixture
def free_tcp_port(free_tcp_port_factory: Callable[[], int]) -> int:
    return free_tcp_port_factory()


@pytest.fixture
def free_udp_port(free_udp_port_factory: Callable[[], int]) -> int:
    return free_udp_port_factory()



from __future__ import annotations

from typing import TYPE_CHECKING

from ._lazyimport import (
    fix_package_names,
    install_lazy_importer,
    set_deprecated_aliases,
)

if TYPE_CHECKING or not install_lazy_importer():
    from ._core._concurrency_utils import amap as amap
    from ._core._concurrency_utils import as_completed as as_completed
    from ._core._concurrency_utils import gather as gather
    from ._core._contextmanagers import (
        AsyncContextManagerMixin as AsyncContextManagerMixin,
    )
    from ._core._contextmanagers import ContextManagerMixin as ContextManagerMixin
    from ._core._eventloop import current_time as current_time
    from ._core._eventloop import get_all_backends as get_all_backends
    from ._core._eventloop import get_available_backends as get_available_backends
    from ._core._eventloop import get_cancelled_exc_class as get_cancelled_exc_class
    from ._core._eventloop import run as run
    from ._core._eventloop import sleep as sleep
    from ._core._eventloop import sleep_forever as sleep_forever
    from ._core._eventloop import sleep_until as sleep_until
    from ._core._exceptions import BrokenResourceError as BrokenResourceError
    from ._core._exceptions import BrokenWorkerInterpreter as BrokenWorkerInterpreter
    from ._core._exceptions import BrokenWorkerProcess as BrokenWorkerProcess
    from ._core._exceptions import BusyResourceError as BusyResourceError
    from ._core._exceptions import ClosedResourceError as ClosedResourceError
    from ._core._exceptions import ConnectionFailed as ConnectionFailed
    from ._core._exceptions import DelimiterNotFound as DelimiterNotFound
    from ._core._exceptions import EndOfStream as EndOfStream
    from ._core._exceptions import FutureAlreadyFinished as FutureAlreadyFinished
    from ._core._exceptions import FutureCancelled as FutureCancelled
    from ._core._exceptions import FutureFailed as FutureFailed
    from ._core._exceptions import FutureNotFinished as FutureNotFinished
    from ._core._exceptions import IncompleteRead as IncompleteRead
    from ._core._exceptions import NoEventLoopError as NoEventLoopError
    from ._core._exceptions import RunFinishedError as RunFinishedError
    from ._core._exceptions import TaskCancelled as TaskCancelled
    from ._core._exceptions import TaskFailed as TaskFailed
    from ._core._exceptions import TaskNotFinished as TaskNotFinished
    from ._core._exceptions import (
        TypedAttributeLookupError as TypedAttributeLookupError,
    )
    from ._core._exceptions import WouldBlock as WouldBlock
    from ._core._fileio import AsyncFile as AsyncFile
    from ._core._fileio import Path as Path
    from ._core._fileio import open_file as open_file
    from ._core._fileio import wrap_file as wrap_file
    from ._core._futures import Future as Future
    from ._core._resources import aclose_forcefully as aclose_forcefully
    from ._core._signals import open_signal_receiver as open_signal_receiver
    from ._core._sockets import TCPConnectable as TCPConnectable
    from ._core._sockets import UNIXConnectable as UNIXConnectable
    from ._core._sockets import as_connectable as as_connectable
    from ._core._sockets import connect_tcp as connect_tcp
    from ._core._sockets import connect_unix as connect_unix
    from ._core._sockets import (
        create_connected_udp_socket as create_connected_udp_socket,
    )
    from ._core._sockets import (
        create_connected_unix_datagram_socket as create_connected_unix_datagram_socket,
    )
    from ._core._sockets import create_tcp_listener as create_tcp_listener
    from ._core._sockets import create_udp_socket as create_udp_socket
    from ._core._sockets import (
        create_unix_datagram_socket as create_unix_datagram_socket,
    )
    from ._core._sockets import create_unix_listener as create_unix_listener
    from ._core._sockets import getaddrinfo as getaddrinfo
    from ._core._sockets import getnameinfo as getnameinfo
    from ._core._sockets import notify_closing as notify_closing
    from ._core._sockets import wait_readable as wait_readable
    from ._core._sockets import wait_socket_readable as wait_socket_readable
    from ._core._sockets import wait_socket_writable as wait_socket_writable
    from ._core._sockets import wait_writable as wait_writable
    from ._core._streams import (
        create_memory_object_stream as create_memory_object_stream,
    )
    from ._core._subprocesses import open_process as open_process
    from ._core._subprocesses import run_process as run_process
    from ._core._synchronization import CapacityLimiter as CapacityLimiter
    from ._core._synchronization import (
        CapacityLimiterStatistics as CapacityLimiterStatistics,
    )
    from ._core._synchronization import Condition as Condition
    from ._core._synchronization import ConditionStatistics as ConditionStatistics
    from ._core._synchronization import Event as Event
    from ._core._synchronization import EventStatistics as EventStatistics
    from ._core._synchronization import Lock as Lock
    from ._core._synchronization import LockStatistics as LockStatistics
    from ._core._synchronization import ResourceGuard as ResourceGuard
    from ._core._synchronization import Semaphore as Semaphore
    from ._core._synchronization import SemaphoreStatistics as SemaphoreStatistics
    from ._core._tasks import TASK_STATUS_IGNORED as TASK_STATUS_IGNORED
    from ._core._tasks import CancelScope as CancelScope
    from ._core._tasks import TaskHandle as TaskHandle
    from ._core._tasks import create_task_group as create_task_group
    from ._core._tasks import current_effective_deadline as current_effective_deadline
    from ._core._tasks import fail_after as fail_after
    from ._core._tasks import fail_at as fail_at
    from ._core._tasks import move_on_after as move_on_after
    from ._core._tasks import move_on_at as move_on_at
    from ._core._tempfile import NamedTemporaryFile as NamedTemporaryFile
    from ._core._tempfile import SpooledTemporaryFile as SpooledTemporaryFile
    from ._core._tempfile import TemporaryDirectory as TemporaryDirectory
    from ._core._tempfile import TemporaryFile as TemporaryFile
    from ._core._tempfile import gettempdir as gettempdir
    from ._core._tempfile import gettempdirb as gettempdirb
    from ._core._tempfile import mkdtemp as mkdtemp
    from ._core._tempfile import mkstemp as mkstemp
    from ._core._testing import TaskInfo as TaskInfo
    from ._core._testing import get_current_task as get_current_task
    from ._core._testing import get_running_tasks as get_running_tasks
    from ._core._testing import wait_all_tasks_blocked as wait_all_tasks_blocked
    from ._core._typedattr import TypedAttributeProvider as TypedAttributeProvider
    from ._core._typedattr import TypedAttributeSet as TypedAttributeSet
    from ._core._typedattr import typed_attribute as typed_attribute

    fix_package_names()
    set_deprecated_aliases(
        {
            "BrokenWorkerIntepreter": "anyio.BrokenWorkerInterpreter",
        }
    )



from __future__ import annotations

from inspect import iscoroutine

__all__ = ("amap", "as_completed", "gather")

from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Iterable
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, TypeVar, overload

from ..abc._tasks import get_coro_name
from ._exceptions import BrokenResourceError
from ._streams import create_memory_object_stream
from ._tasks import TaskHandle, create_task_group

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

R = TypeVar("R")
S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")


@overload
async def gather(
    coro1: Coroutine[Any, Any, R], coro2: Coroutine[Any, Any, S], /
) -> tuple[R, S]: ...


@overload
async def gather(
    coro1: Coroutine[Any, Any, R],
    coro2: Coroutine[Any, Any, S],
    coro3: Coroutine[Any, Any, T],
    /,
) -> tuple[R, S, T]: ...


@overload
async def gather(
    coro1: Coroutine[Any, Any, R],
    coro2: Coroutine[Any, Any, S],
    coro3: Coroutine[Any, Any, T],
    coro4: Coroutine[Any, Any, U],
    /,
) -> tuple[R, S, T, U]: ...


@overload
async def gather(
    coro1: Coroutine[Any, Any, R],
    coro2: Coroutine[Any, Any, S],
    coro3: Coroutine[Any, Any, T],
    coro4: Coroutine[Any, Any, U],
    coro5: Coroutine[Any, Any, V],
    /,
) -> tuple[R, S, T, U, V]: ...


@overload
async def gather(
    coro1: Coroutine[Any, Any, R],
    coro2: Coroutine[Any, Any, S],
    coro3: Coroutine[Any, Any, T],
    coro4: Coroutine[Any, Any, U],
    coro5: Coroutine[Any, Any, V],
    coro6: Coroutine[Any, Any, W],
    /,
) -> tuple[R, S, T, U, V, W]: ...


# handle arbitrary length if awaitables are all of the same type
@overload
async def gather(*coros: Coroutine[Any, Any, R]) -> tuple[R, ...]: ...


async def gather(*coros: Coroutine[Any, Any, Any]) -> tuple[Any, ...]:
    """
    Run coroutines concurrently in a task group. The order of result values corresponds
    to the order of coroutines passed.

    :param coros: coroutine objects to run as tasks
    :return: task results for each argument in the same order as they were passed

    """
    handles: list[TaskHandle[Any]] = []
    async with create_task_group() as tg:
        handles.extend(tg.create_task(coro) for coro in coros)

    return tuple(h.return_value for h in handles)


@asynccontextmanager
async def as_completed(
    *awaitables: Awaitable[R],
) -> AsyncGenerator[MemoryObjectReceiveStream[TaskHandle[R]]]:
    """
    Run awaitable objects concurrently in a task group, returning an iterator which can
    be used to get finished task handles in the order they complete.

    :param awaitables: awaitable objects to run as tasks
    :return: MemoryObjectReceiveStream for iterating over task handles as tasks complete.

    """
    if not awaitables:
        raise ValueError("as_completed() takes at least one awaitable")

    send, recv = create_memory_object_stream[TaskHandle[R]](len(awaitables))
    task_handles: list[TaskHandle[R]] = []

    async def runner(
        awaitable: Awaitable[R],
        index: int,
        _send: MemoryObjectSendStream[TaskHandle[R]],
    ) -> R:
        async with _send:
            try:
                return await awaitable
            finally:
                try:
                    _send.send_nowait(task_handles[index])
                except BrokenResourceError:
                    pass

    async with recv, create_task_group() as tg:
        async with send:
            for i, awaitable in enumerate(awaitables):
                name = get_coro_name(awaitable) if iscoroutine(awaitable) else None
                task_handles.append(
                    tg.start_soon(runner, awaitable, i, send.clone(), name=name)
                )
        try:
            yield recv
        finally:
            tg.cancel_scope.cancel()


async def amap(
    func: Callable[[T], Coroutine[Any, Any, R]], args: Iterable[T]
) -> list[R]:
    """
    Run the given coroutine function concurrently for multiple argument values.

    :param func: a coroutine function that takes a single argument
    :param args: a sequence of argument values to pass to ``func``
    :return: task results for each argument in the same order as they were passed

    """
    handles: list[TaskHandle[R]] = []
    async with create_task_group() as tg:
        handles.extend(tg.start_soon(func, arg) for arg in args)

    return [h.return_value for h in handles]



from __future__ import annotations

from collections.abc import Generator
from enum import Enum, auto
from typing import Any, Generic, TypeVar

from ._exceptions import (
    FutureAlreadyFinished,
    FutureCancelled,
    FutureFailed,
    FutureNotFinished,
)
from ._synchronization import Event

T = TypeVar("T")


class Future(Generic[T]):
    """
    An awaitable object that works similar to a :class:`asyncio.Future` but
    with similar characteristics to a :class:`.TaskHandle`.
    """

    class Status(Enum):
        """
        The status of a future handle.

        .. attribute:: PENDING

            The future has not finished yet.
        .. attribute:: FINISHED

            The future has finished with a return value.
        .. attribute:: CANCELLED

            The future was cancelled and has finished since.
        .. attribute:: FAILED

            The future raised an exception.
        """

        PENDING = auto()
        FINISHED = auto()
        CANCELLED = auto()
        FAILED = auto()

    __slots__ = (
        "_cancelled",
        "_exception",
        "_finished_event",
        "_name",
        "_return_value",
    )
    _return_value: T

    def __init__(self, *, name: str | None = None) -> None:
        self._finished_event = Event()
        self._exception: BaseException | None = None
        self._cancelled: bool = False
        self._name = name

    def _check_pending(self) -> None:
        """Shortcut for checking if a Future is pending

        :raises FutureAlreadyFinished: if future was already given a result or exception.
        :raises FutureCancelled: if future has been cancelled previously.
        """
        match self.status:
            case Future.Status.PENDING:
                return
            case Future.Status.FINISHED:
                raise FutureAlreadyFinished("future has already finished")
            case Future.Status.FAILED:
                raise FutureAlreadyFinished("future already failed")
            case Future.Status.CANCELLED:
                raise FutureCancelled("future was cancelled")

    async def wait(self) -> None:
        """
        Waits for the future to finish.

        This method will attempt to wait for a result or exception
        """
        await self._finished_event.wait()

    def cancel(self) -> None:
        """Cancels a pending `.Future` object

        Does nothing if the Future was already finished.
        """
        if self.status is Future.Status.PENDING:
            self._cancelled = True
            self._finished_event.set()

    @property
    def exception(self) -> BaseException | None:
        """
        The exception value of a `.Future`

        :raises FutureNotFinished: if future is still pending
        :raises FutureCancelled: if future was cancelled
        :returns: None if future succeeds with a result sent instead
            otherwise this will be an exception
        """

        match self.status:
            case Future.Status.PENDING:
                raise FutureNotFinished("the future has not finished yet")
            case Future.Status.FINISHED:
                return None
            case Future.Status.CANCELLED:
                raise FutureCancelled("the future was cancelled")
            case Future.Status.FAILED:
                return self._exception

    @exception.setter
    def exception(self, exception: BaseException) -> None:
        """Send exception for a `.Future` object

        :raises FutureAlreadyFinished: if future was already given a result or exception.
        :raises FutureCancelled: if future has been cancelled previously.
        """
        self._check_pending()
        self._exception = exception
        self._finished_event.set()

    @property
    def return_value(self) -> T:
        """
        The return value of the future.

        :raises FutureNotFinished: if the future has not finished yet
        :raises FutureCancelled: if the future was cancelled
        :raises FutureFailed: if the future raised an exception

        """
        match self.status:
            case Future.Status.PENDING:
                raise FutureNotFinished("the future has not finished yet")
            case Future.Status.FINISHED:
                return self._return_value
            case Future.Status.CANCELLED:
                raise FutureCancelled("the future was cancelled")
            case Future.Status.FAILED:
                raise FutureFailed(
                    "the future raised an exception"
                ) from self._exception

    @return_value.setter
    def return_value(self, value: T) -> None:
        """
        Send pending result for a `.Future` object

        :raises FutureAlreadyFinished: if future was already given a result or exception.
        :raises FutureCancelled: if future has been cancelled previously.
        """
        self._check_pending()
        self._return_value = value
        self._finished_event.set()

    @property
    def status(self) -> Future.Status:
        """
        The current status of a future.

        Every future starts in the :attr:`~Future.Status.PENDING` state.
        If a future is cancelled while in this state, it will transition to the
        :attr:`Future.Status.CANCELLED` state. When the task finishes, it will
        transition to one of the three final states (
        :attr:`Future.Status.FINISHED`, :attr:`Future.Status.FAILED`, or
        :attr:`Future.Status.CANCELLED`) depending on the exception the task
        raised, if any. No other status transitions will happen.
        """
        if not self._finished_event.is_set():
            return Future.Status.PENDING
        elif self._cancelled:
            return Future.Status.CANCELLED
        elif self._exception is not None:
            return Future.Status.FAILED
        else:
            return Future.Status.FINISHED

    def __await__(self) -> Generator[Any, Any, T]:
        yield from self.wait().__await__()
        return self.return_value

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} {self.status.name.lower()} "
            f"name={self._name!r}>"
        )
